searchData={"items":[{"type":"module","title":"FDBC","doc":"An unofficial FoundationDB client for Elixir.\n\nFoundationDB is a distributed database designed to handle large volumes of\nstructured data across clusters of commodity servers. It organises data as an\nordered key-value store and employs ACID transactions for all operations. It\nis especially well-suited for read/write workloads but also has excellent\nperformance for write-intensive workloads.\n\nThis documentation makes use of and refers to the [upstream\ndocumentation](https://apple.github.io/foundationdb/index.html) where it\nmakes sense.","ref":"FDBC.html"},{"type":"module","title":"Foundation - FDBC","doc":"`FDBC.Foundation` is a wrapper around the database geared at easing use within applications.\n\n```elixir\ndefmodule Demo.Foundation do\n  use FDBC.Foundation,\n    otp_app: :demo,\n    options: [\n      transaction_retry_limit: 100,\n      transaction_timeout: 60_000\n    ]\nend\n```\n\nWhere the configuration for the Foundation must be in your application\nenvironment, usually defined in your `config/config.exs`:\n\n```elixir\nconfig :demo, Demo.Foundation,\n  cluster: \"./fdb.cluster\",\n  directory: \"demo\"\n```\n\nIf your application was generated with a supervisor you will have a\n`lib/demo/application.ex` file containing the application start callback that\ndefines and starts your supervisor. You just need to edit the `start/2`\nfunction to start the repo as a supervisor on your application's supervisor:\n\n```elixir\ndef start(_type, _args) do\n  children = [\n    Demo.Foundation,\n  ]\n\n  opts = [strategy: :one_for_one, name: Demo.Supervisor]\n  Supervisor.start_link(children, opts)\nend\n```","ref":"FDBC.html#module-foundation"},{"type":"function","title":"FDBC.api_version/1","doc":"Specifies the version of the API that the application uses. This allows\nfuture versions of FoundationDB to make API changes without breaking existing\nprograms.\n\nIf the version is not specified the latest currently supported API version is used.\n\n> #### Note {: .info}\n> This function must be called before any other part of this library, unless\n  being used as part of an OTP application.","ref":"FDBC.html#api_version/1"},{"type":"function","title":"FDBC.transact/3","doc":"Performs a transaction against the database.\n\nThis is a convenience funtion that will create a transaction, passing it to\nthe provided function, committing it and handling any retry logic based on\nthe error returned.\n\nFurthermore a transaction can be passed to this function in which case it\nlogically does nothing, this is useful to allow composition of transactional\nfunctions.\n\nThe options are used for creating the transaction and therefore are the same\nas those for `FDBC.Transaction.create/2`.","ref":"FDBC.html#transact/3"},{"type":"module","title":"FDBC.Database","doc":"An `FDBC.Database` represents a FoundationDB database — a mutable,\nlexicographically ordered mapping from binary keys to binary values.\nModifications to a database are performed via transactions.","ref":"FDBC.Database.html"},{"type":"function","title":"FDBC.Database.change/2","doc":"Change the options on the database instance after its initial creation.\n\nSee `create/2` for the list of options.","ref":"FDBC.Database.html#change/2"},{"type":"function","title":"FDBC.Database.create/2","doc":"Creates a database instance used to interact with the cluster.\n\nIf no cluster file is provided, then FoundationDB automatically [determines a\ncluster file](https://apple.github.io/foundationdb/administration.html#specifying-a-cluster-file)\nwith which to connect to a cluster.\n\nA single client can use this function multiple times to connect to different\nclusters simultaneously, with each invocation requiring its own cluster file.","ref":"FDBC.Database.html#create/2"},{"type":"function","title":"Options - FDBC.Database.create/2","doc":"* `:datacenter_id` - (binary) Specify the datacenter ID that was passed to\n    fdbserver processes running in the same datacenter as this client, for\n    better location-aware load balancing.\n\n  * `:location_cache_size` - (integer) Set the size of the client location\n    cache. Raising this value can boost performance in very large databases\n    where clients acc  ess data in a near-random pattern. Defaults to `100000`.\n\n  * `:machine_id` - (binary) Specify the machine ID that was passed to\n    fdbserver processes running on the same machine as this client, for\n    better location-aware load balancing.\n\n  * `:max_watches` - (integer) Set the maximum number of watches allowed to be\n    outstanding on a database connection. Increasing this number could result\n    in increased resource usage. Reducing this number will not cancel any\n    outstanding watches. Defaults to `10000` and cannot be larger than `1000000`.\n\n  * `:snapshot_ryw` - (boolean) Specify whether snapshot read operations will\n    see the results of writes done in the same transaction. Defaults to `true`.\n\n  * `:test_causal_read_risky` - (integer) Enables verification of causal read\n    risky by checking whether clients are able to read stale data when they\n    detect a recovery, and logging an error if so. Must be between 0 and 100\n    expressing the probability a client will verify it can't read stale data.\n\n  * `:transaction_automatic_idempotency` - (true) Set a random idempotency id\n    for all transactions. See the transaction option description for more\n    information. This feature is in development and not ready for general\n    use.\n\n  * `:transaction_bypass_unreadable` - (true) Allows ``get`` operations to\n    read from sections of keyspace that have become unreadable because of\n    versionstamp operations. This sets the ``bypass_unreadable`` option of\n    each transaction created by this database. See the transaction option\n    description for more information.\n\n  * `:transaction_causal_read_risky` - (true) The read version will be\n    committed, and usually will be the latest committed, but might not be the\n    latest committed in the event of a simultaneous fault and misbehaving\n    clock.\n\n  * `:transaction_logging_max_field_length` - (integer) Sets the maximum\n    escaped length of key and value fields to be logged to the trace file via\n    the `:log_transaction` option. This sets the `:transaction_logging_max_field_length`\n    option of each transaction created by this database. See the transaction\n    option description for more information.\n\n  * `:transaction_max_retry_delay` - (integer) Set the maximum amount of\n    backoff delay incurred in the call to `on_error` if the error is retryable.\n    This sets the `:max_retry_delay` option of each transaction created by\n    this database. See the transaction option description for more information.\n\n  * `:transaction_report_conflicting_keys` - (true) Enables conflicting key\n    reporting on all transactions, allowing them to retrieve the keys that\n    are conflicting with other transactions.\n\n  * `:transaction_retry_limit` - (integer) Set a maximum number of retries\n    after which additional calls to `on_error` will throw the most recently\n    seen error code. This sets the `:retry_limit` option of each transaction\n    created by this database. See the transaction option description for more\n    information.\n\n  * `:transaction_size_limit` - (integer) Set the maximum transaction size in\n    bytes. This sets the `:size_limit` option on each transaction created by\n    this database. See the transaction option description for more information.\n\n  * `:transaction_timeout` - (integer) Set a timeout in milliseconds which,\n    when elapsed, will cause each transaction automatically to be cancelled.\n    This sets the `:timeout` option of each transaction created by this\n    database. See the transaction option description for more information.\n    Using this option requires that the API version is 610 or higher.\n\n  * `:transaction_used_during_commit_protection_disable` - (true) By default,\n    operations that are performed on a transaction while it is being\n    committed will not only fail themselves, but they will attempt to fail\n    other in-flight operations (such as the commit) as well. This behavior is\n    intended to help developers discover situations where operations could be\n    unintentionally executed after the transaction has been reset. Setting\n    this option removes that protection, causing only the offending operation\n    to fail.\n\n  * `:use_config_database` - (true) Use configuration database.","ref":"FDBC.Database.html#create/2-options"},{"type":"type","title":"FDBC.Database.t/0","doc":"","ref":"FDBC.Database.html#t:t/0"},{"type":"module","title":"FDBC.Directory","doc":"FoundationDB provides the directories layer as a tool for managing related\nsubspaces. Directories are a recommended approach for administering\napplications. Each application should create or open at least one directory\nto manage its subspaces.\n\nDirectories are identified by hierarchical paths analogous to the paths in a\nUnix-like file system. A path is represented as a tuple of strings. Each\ndirectory has an associated subspace used to store its content. The directory\nlayer maps each path to a short prefix used for the corresponding subspace.\nIn effect, directories provide a level of indirection for access to\nsubspaces.\n\nThis design has significant benefits: while directories are logically\nhierarchical as represented by their paths, their subspaces are not\nphysically nested in a corresponding way. For example, suppose we create a\nfew directories with increasing paths, such as:\n\n```elixir\niex> r = FDBC.Directory.new()\niex> a = FDBC.Directory.create!(r, db, ['alpha'])\niex> b = FDBC.Directory.create!(r, db, ['alpha', 'bravo'])\niex> c = FDBC.Directory.create!(r, db, ['alpha', 'bravo', 'charlie'])\n```\n\nThe prefixes of `a`, `b`, and `c` are allocated independently and will\nusually not increase in length. The indirection from paths to subspaces\nallows keys to be kept short and makes it fast to move directories (i.e.,\nrename their paths).\n\nPaths in the directory layer are always relative. In particular, paths are\ninterpreted relative to the directory in which an operation is performed. For\nexample, we could have created the directories `a`, `b`, and `c` as follows:\n\n```elixir\niex> r = FDBC.Directory.new()\niex> a = FDBC.Directory.open!(r, db, ['alpha'], create: true)\niex> b = FDBC.Directory.open!(a, db, ['bravo'], create: true)\niex> c = FDBC.Directory.open!(b, db, ['charlie'], create: true)\n```","ref":"FDBC.Directory.html"},{"type":"module","title":"Usage - FDBC.Directory","doc":"Unlike the upstream implementations this one tries to remain closer to the\nanalogy of a 'Unix-like' file system. By default `new/1` will create default\nroot directory which is that with a content subspace of `<<>>` and a metadata\nprefix of `<<0xFE>>`.\n\n> #### Info {: .info}\n> When comparing to upstream documentation what its refers to as\n`node_subspace` is called `metadata` here.\n\nThe `create/3` function will create a new directory but not return it, to get\nthe newly created directory in the same request `open/3` should be used:\n\n```elixir\niex> :ok = FDBC.Directory.new() |> FDBC.Directory.create!(db, ['users'])\niex> users_dir = FDBC.Directory.new() |> FDBC.Directory.open!(db, ['users'], create: true)\n```\n\nA directory that is not a partition exposes its `FDBC.Subspace` for usage.\nThe root directory is a partition.\n\n```elixir\niex> r = FDBC.Directory.new()\niex> nil = FDBC.Directory.subspace(r)\niex> users_dir = FDBC.Directory.open!(r, db, [\"users\"], create: true)\niex> %FDBC.Subspace{} = FDBC.Directory.subspace(users_dir)\n```\n\nIf the directory was created previously (e.g., in a prior session or by\nanother client), you can open it via its path:\n\n```elixir\niex> {:ok, users} = FDBC.Directory.new() |> FDBC.Directory.open(db, ['users'])\n```\n\nSimilarly to a file system directories can be enumerated, moved and removed,\nwhere all paths are relative to the directory provided:\n\n```elixir\niex> subdirs = FDBC.Directory.new() |> FDBC.Directory.list(db, [])\niex> FDBC.Directory.new() |> FDBC.Directory.move(db, [\"a\"], [\"b\"])\niex> FDBC.Directory.new() |> FDBC.Directory.remove(db, [\"a\"])\n```","ref":"FDBC.Directory.html#module-usage"},{"type":"module","title":"Partitions - FDBC.Directory","doc":"Under normal operation, a directory does not share a common key prefix with\nits subdirectories. As a result, related directories are not necessarily\nlocated together in key-space. This means that you cannot use a single range\nquery to read all contents of a directory and its descendants simultaneously,\nfor example.\n\nFor most applications this behavior is acceptable, but in some cases it may\nbe useful to have a directory tree in your hierarchy where every directory\nshares a common prefix. For this purpose, the directory layer supports\ncreating partitions within a directory. A partition is a directory whose\nprefix is prepended to all of its descendant directories’ prefixes.\n\nA partition can be created by setting `:partition` to in options, this will\nalso set the directories label to `partition` and thus trying to create a\npartition with a custom label will raise an exception.\n\n```elixir\niex> r = FDBC.Directory.new()\niex> FDBC.Directory.create(r, db, [\"p1\"], partition: true)\niex> users_dir = FDBC.Directory.open!(r, db, [\"p1\", \"users\"], create: true)\n```\n\nDirectory partitions have the following drawbacks, and in general they should\nnot be used unless specifically needed:\n\n* Directories cannot be moved between different partitions.\n\n* Directories in a partition have longer prefixes than their counterparts\noutside of partitions, which reduces performance. Nesting partitions inside\nof other partitions results in even longer prefixes.\n\n* The root directory of a partition cannot be used to pack/unpack keys and\ntherefore cannot be used to create subspaces. You must create at least one\nsubdirectory of a partition in order to store content in it.","ref":"FDBC.Directory.html#module-partitions"},{"type":"function","title":"FDBC.Directory.create/4","doc":"Creates a directory at the given path relative to the current working directory.","ref":"FDBC.Directory.html#create/4"},{"type":"function","title":"Options - FDBC.Directory.create/4","doc":"* `:label` - the label to assign to the directory to aid in its identification.\n\n  * `:parents` - allows the creation of parent directories if not present,\n    otherwise an exception will be raised.\n\n  * `:partition` - the directory is created as a partition whose prefix is\n    prepended to all of its descendant directories’ prefixes.\n\n  * `:prefix` - creates the directory with the provided prefix, otherwise\n    the prefix is allocated automatically.","ref":"FDBC.Directory.html#create/4-options"},{"type":"function","title":"FDBC.Directory.create!/4","doc":"Similar to `create/4` but raises an error on failure.","ref":"FDBC.Directory.html#create!/4"},{"type":"function","title":"FDBC.Directory.exists?/3","doc":"Checks if a directory exists for the given path relative to the current\nworking directory.\n\nIf no path is provided then the current working directory is checked.","ref":"FDBC.Directory.html#exists?/3"},{"type":"function","title":"FDBC.Directory.label/1","doc":"Returns the label of the provided directory.","ref":"FDBC.Directory.html#label/1"},{"type":"function","title":"FDBC.Directory.list/3","doc":"List the immediate subdirectories for the given path relative to the current\nworking directory.","ref":"FDBC.Directory.html#list/3"},{"type":"function","title":"FDBC.Directory.list!/3","doc":"Similar to `list/3` but raises an error on failure.","ref":"FDBC.Directory.html#list!/3"},{"type":"function","title":"FDBC.Directory.move/5","doc":"Move the `source` directory to the `destination` directory, relative to the\ncurrent working directory.","ref":"FDBC.Directory.html#move/5"},{"type":"function","title":"Options - FDBC.Directory.move/5","doc":"* `:parents` - allows the creation of parent directories if not present,\n    otherwise an exception will be raised.","ref":"FDBC.Directory.html#move/5-options"},{"type":"function","title":"FDBC.Directory.move!/5","doc":"Similar to `move/4` but raises an error on failure.","ref":"FDBC.Directory.html#move!/5"},{"type":"function","title":"FDBC.Directory.new/1","doc":"The default instance of the root directory.","ref":"FDBC.Directory.html#new/1"},{"type":"function","title":"Options - FDBC.Directory.new/1","doc":"* `:allow_manual_prefixes` - allows the creation of manual prefixes, if not\n    enabled attempts to provide a manual prefix will result in an exception.\n\n  * `:partition` - allows for a different partition other than the root to be\n    used. By default an empty prefix is used for the partition's content\n    subspace, while the value `0xFE` is used for the metadata subspace.","ref":"FDBC.Directory.html#new/1-options"},{"type":"function","title":"FDBC.Directory.open/4","doc":"Opens the directory at the given path relative to the current working directory.","ref":"FDBC.Directory.html#open/4"},{"type":"function","title":"Options - FDBC.Directory.open/4","doc":"* `:create` - allow the directory to be created if it does not exist,\n    in which case the options from `create/3` will also be honoured.","ref":"FDBC.Directory.html#open/4-options"},{"type":"function","title":"FDBC.Directory.open!/4","doc":"Similar to `open/4` but raises an error on failure.","ref":"FDBC.Directory.html#open!/4"},{"type":"type","title":"FDBC.Directory.partition/0","doc":"","ref":"FDBC.Directory.html#t:partition/0"},{"type":"function","title":"FDBC.Directory.path/1","doc":"Returns the path of the provided directory.","ref":"FDBC.Directory.html#path/1"},{"type":"function","title":"FDBC.Directory.remove/3","doc":"Remove the directory at the given path relative to the current working\ndirectory.","ref":"FDBC.Directory.html#remove/3"},{"type":"function","title":"FDBC.Directory.remove!/3","doc":"Similar to `remove/3` but raises an error on failure.","ref":"FDBC.Directory.html#remove!/3"},{"type":"function","title":"FDBC.Directory.subspace/1","doc":"Returns the `Subspace.t()` of the provided directory, or `nil` if the\ndirectory is a partition.","ref":"FDBC.Directory.html#subspace/1"},{"type":"function","title":"FDBC.Directory.subspace!/1","doc":"Similar to `subspace/1` but raises an error on failure.","ref":"FDBC.Directory.html#subspace!/1"},{"type":"type","title":"FDBC.Directory.t/0","doc":"","ref":"FDBC.Directory.html#t:t/0"},{"type":"behaviour","title":"FDBC.Foundation","doc":"A convenience behaviour to ease the use of a database instance within an OTP\napplication while also promoting best practices.\n\nFirstly the configuration for the behaviour must be set and is designed in\nsuch a way that it can be changed for each environment:\n\n```elixir\nconfig :example, Example.Foundation,\n  cluster: \"./fdb.cluster\",  # Path to the cluster file\n  directory: \"example\",      # The directory the application should operate under\n```\n\nThe behaviour can then be used to create and manage a database instance.\n\n```elixir\ndefmodule Example.Foundation do\n  use FDBC.Foundation,\n    otp_app: :example,\n    options: [\n      transaction_retry_limit: 100,\n      transaction_timeout: 60_000\n    ]\nend\n```\n\nThe module must be added to the applications supervision tree, once this is\ndone it can then be used to easily transact with FoundationDB.\n\n```elixir\ndefmodule Example.App do\n  alias FDBC.{Directory, Subspace, Transaction}\n  alias Example.Foundation\n\n  def get_value(key) do\n    Foundation.transact(fn tr, dir ->\n      key = Directory.subspace(dir) |> Subspace.pack([key])\n      Transaction.get(tr, key)\n    end)\n  end\n\n  def set_value(key, value) do\n    Foundation.transact(fn tr, dir ->\n      key = Directory.subspace(dir) |> Subspace.pack([key])\n      Transaction.set(tr, key, value)\n    end)\n  end\nend\n```","ref":"FDBC.Foundation.html"},{"type":"callback","title":"FDBC.Foundation.database/0","doc":"Returns the underlying database instance.","ref":"FDBC.Foundation.html#c:database/0"},{"type":"callback","title":"FDBC.Foundation.transact/2","doc":"Performs a transaction against the database.\n\nThe directory passed to `fun` is that which namespaces the application as\ndefined in the configuration.\n\nThe options are used for creating the transaction and therefore are the same\nas those for `FDBC.Transaction.create/2`.","ref":"FDBC.Foundation.html#c:transact/2"},{"type":"module","title":"FDBC.Future","doc":"Most functions in the FoundationDB API are asynchronous, meaning that they\nmay return to the caller before actually delivering their result. An\n`FDBC.Future` object represents a result value or error to be delivered at\nsome future time. A future can be awaited on or periodically checked to see\nif it is ready. Once a Future is ready it can be resolved to get its\nunderlying value, which is dependant on the function that created the future.\n\nSee `FDBC.Transaction` for examples on how futures can be used.","ref":"FDBC.Future.html"},{"type":"function","title":"FDBC.Future.await/2","doc":"Awaits a future returning the result for its given type.\n\nThis is similar to `resolve/1` except that a timeout can be set.\n\nOn error it raises a `FDBC.Error` exception.","ref":"FDBC.Future.html#await/2"},{"type":"function","title":"FDBC.Future.await_many/2","doc":"Awaits multiple futures returning their results.\n\nThis function receives a list of futures and waits for their results in the\ngiven time interval. It returns a list of the results, in the same order as\nthe futures supplied in the `futures` input argument.\n\nA timeout, in milliseconds or `:infinity`, can be given with a default value\nof `5000`. If the timeout is exceeded, then a `FDBC.Error` with a value of\ntimed out is raised.\n\nOn error it raises a `FDBC.Error` exception.","ref":"FDBC.Future.html#await_many/2"},{"type":"function","title":"FDBC.Future.cancel/1","doc":"Cancels a future and its associated asynchronous operation.\n\nIf called before the future is ready, attempts to `resolve/1` the future will\nresult in a `FDBC.Error` exception with\n[operation_cancelled](https://apple.github.io/foundationdb/api-error-codes.html#developer-guide-error-codes)\nerror.\n\nCancelling a future which is already ready has no effect.\n\n> #### Note {: .info}\n> Even if a future is not ready, its associated asynchronous operation may\n  have succesfully completed and be unable to be cancelled.","ref":"FDBC.Future.html#cancel/1"},{"type":"function","title":"FDBC.Future.ready?/1","doc":"Returns `true` if the future is ready, where a Future is ready if it has been\nset to a value or an error.","ref":"FDBC.Future.html#ready?/1"},{"type":"function","title":"FDBC.Future.resolve/1","doc":"Resolves a future returning the result for its given type.\n\nOn error it raises a `FDBC.Error` exception.","ref":"FDBC.Future.html#resolve/1"},{"type":"type","title":"FDBC.Future.t/0","doc":"","ref":"FDBC.Future.html#t:t/0"},{"type":"type","title":"FDBC.Future.t/1","doc":"","ref":"FDBC.Future.html#t:t/1"},{"type":"module","title":"FDBC.KeySelector","doc":"FoundationDB’s lexicographically ordered data model permits finding keys\nbased on their order (for example, finding the first key in the database\ngreater than a given key). Key selectors represent a description of a key in\nthe database that could be resolved to an actual key by\n`FDBC.Transaction.get_key/3` or used directly as the beginning or end of a\nrange in `FDBC.Transaction.get_range/4`.\n\nFor example:\n\n```elixir\n# Selects the first key following apple.\nFDBC.KeySelector.greater_than('apple')\n\n# Selects the second key following apple.\nselector = FDBC.KeySelector.greater_than('apple')\n%{selector | selector.offset + 1}\n```\n\n> #### Note {: .info}\n> The current version of FoundationDB does not resolve key selectors with\nlarge offsets in O(1) time. See the [Key selectors with large offsets are\nslow](https://apple.github.io/foundationdb/known-limitations.html#dont-use-key-selectors-for-paging)\nknown limitation.\n\nAll possible key selectors can be constructed using one of the four 'common\nform; constructors and a positive or negative offset. Alternatively, general\nkey selectors can be manually constructed by specifying:\n\n  * A reference key (binary)\n  * An equality flag (boolean)\n  * An offset (integer)\n\nTo 'resolve' these key selectors FoundationDB first finds the last key less\nthan the reference key (or equal to the reference key, if the equality flag\nis true), then moves forward a number of keys equal to the offset (or\nbackwards, if the offset is negative). If a key selector would otherwise\ndescribe a key off the beginning of the database (before the first key), it\ninstead resolves to the empty key `<<>>`. If it would otherwise describe a key\noff the end of the database (after the last key), it instead resolves to the\nkey `'\\xFF'` (or `'\\xFF\\xFF'` if the transaction has been granted access\nto the system keys).","ref":"FDBC.KeySelector.html"},{"type":"function","title":"FDBC.KeySelector.greater_than/1","doc":"Returns a key selector that will select the lexicographically least key\npresent in the database which is lexicographically strictly greater than the\ngiven byte string key.","ref":"FDBC.KeySelector.html#greater_than/1"},{"type":"function","title":"FDBC.KeySelector.greater_than_or_equal/1","doc":"Returns a key selector that will select the lexicographically least key\npresent in the database which is lexicographically greater than or equal to\nthe given byte string key.","ref":"FDBC.KeySelector.html#greater_than_or_equal/1"},{"type":"function","title":"FDBC.KeySelector.less_than/1","doc":"Returns a key selector that will select the lexicographically greatest key\npresent in the database which is lexicographically strictly less than the\ngiven byte string.","ref":"FDBC.KeySelector.html#less_than/1"},{"type":"function","title":"FDBC.KeySelector.less_than_or_equal/1","doc":"Returns a key selector that will select the lexicographically greatest key\npresent in the database which is lexicographically less than or equal to the\ngiven byte string key.","ref":"FDBC.KeySelector.html#less_than_or_equal/1"},{"type":"function","title":"FDBC.KeySelector.new/3","doc":"Returns a key selector for use in key or range based transactions.\n\nThe value of `or_equal` is used to decide whether to include matching the\n`key` itself or just on the key immediate before it.\n\nThe value of `offset` determines which key to select after the match on the `key`.\n\n> #### Note {: .info}\n> FoundationDB does not efficiently resolve key selectors with large offsets,\nso [Key selectors with large offsets are slow](https://apple.github.io/foundationdb/known-limitations.html#dont-use-key-selectors-for-paging).","ref":"FDBC.KeySelector.html#new/3"},{"type":"type","title":"FDBC.KeySelector.t/0","doc":"","ref":"FDBC.KeySelector.html#t:t/0"},{"type":"module","title":"FDBC.Network","doc":"A module used to start and stop the network thread which is responsible for\nrunning most of the clients tasks.","ref":"FDBC.Network.html"},{"type":"function","title":"FDBC.Network.start/1","doc":"Starts the networking thread used for communication with FoundationDB.","ref":"FDBC.Network.html#start/1"},{"type":"function","title":"Options - FDBC.Network.start/1","doc":"* `:callbacks_on_external_threats` - (true) If set, callbacks from external\n    client libraries can be called from threads created by the FoundationDB\n    client library. Otherwise, callbacks will be called from either the\n    thread used to add the callback or the network thread. Setting this\n    option can improve performance when connected using an external client,\n    but may not be safe to use in all environments. Must be set before\n    setting up the network. WARNING: This feature is considered experimental\n    at this time.\n\n  * `:client_threads_per_version` - (integer) Spawns multiple worker threads\n    for each version of the client that is loaded. Setting this to a number\n    greater than one implies `:disable_local_client`.\n\n  * `:client_tmp_dir` - (binary) Sets the directory for storing temporary\n    files created by FDBC client, such as temporary copies of client\n    libraries. Defaults to `/tmp`.\n\n  * `:disable_client_bypass` - (true) Prevents the multi-version client API\n    from being disabled, even if no external clients are configured. This\n    option is required to use GRV caching.\n\n  * `:disable_client_statistics_logging` - (true) Disables logging of client\n    statistics, such as sampled transaction activity.\n\n  * `:disable_local_client` - (true) Prevents connections through the local\n    client, allowing only connections through externally loaded client\n    libraries.\n\n  * `:disable_multi_version_client_api` - (true) Disables the multi-version\n    client API and instead uses the local client directly.\n\n  * `:distributed_client_tracer` - (atom) Set a tracer to run on the client.\n    Should be set to the same value as the tracer set on the server. Valid\n    values are `nil`, `:log_file` or `:network_lossy`.\n\n  * `:enable_run_loop_profiling` - (true) Enables debugging feature to\n    perform run loop profiling. Requires trace logging to be enabled.\n    WARNING: this feature is not recommended for use in production.\n\n  * `:external_client_directory` - (binary) Searches the specified path for\n    dynamic libraries and adds them to the list of client libraries for use\n    by the multi-version client API.\n\n  * `:external_client_library` - (binary) Adds an external client library for\n    use by the multi-version client API.\n\n  * `:fail_incompatible_client` - (true) Fail with an error if there is no\n    client matching the server version the client is connecting to.\n\n  * `:ignore_external_client_failures` - (true) Ignore the failure to\n    initialize some of the external clients.\n\n  * `:knob` - (binary) Set internal tuning or debugging knobs.\n\n  * `:retain_client_library_copies` - (true) Retain temporary external client\n    library copies that are created for enabling multi-threading.\n\n  * `:tls_ca_bytes` - (binary) Set the certificate authority bundle.\n\n  * `:tls_ca_path` - (binary) Set the file from which to load the certificate\n    authority bundle.\n\n  * `:tls_cert_bytes` - (binary) Set the certificate chain.\n\n  * `:tls_cert_path` - (binary) Set the file from which to load the\n    certificate chain.\n\n  * `:tls_disable_plaintext_connection` - (true) Prevent client from\n    connecting to a non-TLS endpoint by throwing network connection failed\n    error.\n\n  * `:tls_key_bytes` - (binary) Set the private key corresponding to your own\n    certificate.\n\n  * `:tls_key_path` - (binary) Set the file from which to load the private\n    key corresponding to your own certificate.\n\n  * `:tls_password` - (binary) Set the passphrase for encrypted private key.\n\n  * `:tls_verify_peers` - (true) Set the peer certificate field verification\n    criteria.\n\n  * `:trace_clock_source` - (atom) Select clock source for trace files. Valid\n    values are `:now` or `realtime`. Defaults to `:now`.\n\n  * `:trace_enable` - (true) Enables trace output to a file in a directory of\n    the clients choosing.\n\n  * `:trace_file_identifier` - (binary) Once provided, this string will be\n    used to replace the port/PID in the log file names.\n\n  * `:trace_format` - (atom) Select the format of the log files. Valid values\n    are `:xml` and `:json`. Defaults to `:xml`.\n\n  * `:trace_initialize_on_setup` - (true) Initialize trace files on network\n    setup, determine the local IP later. Otherwise tracing is initialized when\n    opening the first database.\n\n  * `:trace_log_group` - (binary) Sets the 'LogGroup' attribute with the\n    specified value for all events in the trace output files. The default log\n    group is 'default'.\n\n  * `:trace_max_logs_size` - (integer) Sets the maximum size of all the trace\n    output files put together. If the value is set to 0, there is no limit on\n    the total size of the files. The default is a maximum size of 104,857,600\n    bytes. If the de fault roll size is used, this means that a maximum of 10\n    trace files will be written at a time.\n\n  * `:trace_partial_file_suffix` - (binary) Set file suffix for partially\n    written log files.\n\n  * `:trace_roll_size` - (integer) Sets the maximum size in bytes of a single\n    trace output file. If the value is set to 0, there is no limit on\n    individual file size. The default is a maximum size of 10,485,760 bytes.\n\n  * `:trace_share_among_client_threads` - (true) Use the same base trace file\n    name for all client threads as it did before version 7.2. The current\n    default behavior is to use distinct trace file names for client threads\n    by including their version and thread index.","ref":"FDBC.Network.html#start/1-options"},{"type":"function","title":"FDBC.Network.stop/0","doc":"Stops the network threat.\n\n> #### Warning {: .warning}\n> Once the network is stopped it cannot be restarted during the lifetime of\n  the running program.","ref":"FDBC.Network.html#stop/0"},{"type":"module","title":"FDBC.Subspace","doc":"Subspaces provide the recommended way to define\n[namespaces](https://apple.github.io/foundationdb/developer-guide.html#developer-guide-namespace-management)\nfor different categories of data. As a best practice, you should always use\nat least one subspace as a namespace for your application data.\n\nAn instance of the class stores a prefix used to identify the namespace and\nautomatically prepends it when encoding tuples into keys. Likewise, it\nremoves the prefix when decoding keys.\n\nFor example, suppose your application tracks profile data for your users. You\ncould store the data in a `user_space` subspace that would make `'user'` the\nfirst element of each tuple. A back-end function might have the form:\n\n```elixir\ndefmodule Example do\n  @user_space FDBC.Subspace([\"user\"])\n\n  def set_user_data(db_or_tr, key, value) do\n    FDBC.transact(db_or_tr, fn tr ->\n      FDBC.Transaction.set(tr, FDBC.Subspace.pack(@user_space, [key]), value)\n    end)\n  end\nend\n```","ref":"FDBC.Subspace.html"},{"type":"function","title":"FDBC.Subspace.concat/2","doc":"Returns a new subspace that is the concatination of two subspaces.\n\nThe second subspace is appended to the first subspace.","ref":"FDBC.Subspace.html#concat/2"},{"type":"function","title":"FDBC.Subspace.contains?/2","doc":"Returns true if the subspace is the logical prefix of the given key.","ref":"FDBC.Subspace.html#contains?/2"},{"type":"function","title":"FDBC.Subspace.new/1","doc":"Creates a new subspace with the given prefix.\n\nIf the prefix is a `binary` it is encoded raw, otherwise a tuple must be\nprovided.","ref":"FDBC.Subspace.html#new/1"},{"type":"function","title":"FDBC.Subspace.pack/3","doc":"Returns a key encoding the specified tuple prefixed by the subspace.","ref":"FDBC.Subspace.html#pack/3"},{"type":"function","title":"Options - FDBC.Subspace.pack/3","doc":"* `:extension` - a module that implements the behaviour `FDBC.Tuple.Extension`\n    allowing the Tuple layer to be extended with user type codes.\n\n  * `:keyed` - if present, then the tuple is returned as a keyword list, such\n    as `[{string: \"flip\"}, {string: \"flop\"}]`.","ref":"FDBC.Subspace.html#pack/3-options"},{"type":"function","title":"FDBC.Subspace.range/3","doc":"Returns the start and stop keys in the form of a tuple pair representing all\nkeys in the subspace that encode tuples strictly starting with the specifed\ntuple.","ref":"FDBC.Subspace.html#range/3"},{"type":"type","title":"FDBC.Subspace.t/0","doc":"","ref":"FDBC.Subspace.html#t:t/0"},{"type":"function","title":"FDBC.Subspace.unpack/3","doc":"Returns the tuple that encoded into the given key with the subspace prefix removed.\n\nRaises `ArgumentError` if the `key` is not in the subspace.","ref":"FDBC.Subspace.html#unpack/3"},{"type":"function","title":"Options - FDBC.Subspace.unpack/3","doc":"* `:extension` - a module that implements the behaviour `FDBC.Tuple.Extension`\n    allowing the Tuple layer to be extended with user type codes.\n\n  * `:keyed` - if present, then the tuple is returned as a keyword list, such\n    as `[{string: \"flip\"}, {string: \"flop\"}]`.","ref":"FDBC.Subspace.html#unpack/3-options"},{"type":"module","title":"FDBC.Tenant","doc":"Represents a FoundationDB tenant. Tenants are optional named transaction\ndomains that can be used to provide multiple disjoint key-spaces to client\napplications. A transaction created in a tenant will be limited to the keys\ncontained within that tenant, and transactions operating on different tenants\ncan use the same key names without interfering with each other.","ref":"FDBC.Tenant.html"},{"type":"function","title":"FDBC.Tenant.open/2","doc":"Opens a tenant on the given database.\n\nAll transactions created by this tenant will operate on the tenant’s\nkey-space.","ref":"FDBC.Tenant.html#open/2"},{"type":"type","title":"FDBC.Tenant.t/0","doc":"","ref":"FDBC.Tenant.html#t:t/0"},{"type":"module","title":"FDBC.Transaction","doc":"In FoundationDB, a transaction is a mutable snapshot of a database. All read\nand write operations on a transaction see and modify an otherwise-unchanging\nversion of the database and only change the underlying database if and when\nthe transaction is committed. Read operations do see the effects of previous\nwrite operations on the same transaction. Committing a transaction usually\nsucceeds in the absence of\n[conflicts](https://apple.github.io/foundationdb/developer-guide.html#conflict-ranges).\n\nTransactions group operations into a unit with the properties of atomicity,\nisolation, and durability. Transactions also provide the ability to maintain\nan application’s invariants or integrity constraints, supporting the property\nof consistency. Together these properties are known as\n[ACID](https://apple.github.io/foundationdb/developer-guide.html#acid).\n\nTransactions are also causally consistent: once a transaction has been\nsuccessfully committed, all subsequently created transactions will see the\nmodifications made by it.\n\nApplications must provide error handling and an appropriate retry loop around\nthe application code for a transaction. `FDBC` provides a convenience\nfunction `FDBC.transact/3` which will do just that when passed a\n`FDBC.Database`. The function roughly does the following:\n\n```elixir\ntr = Transaction.create(db)\ndef transact(tr, fun) do\n  result = fun.(transaction)\n  :ok = FDBC.Transaction.commit(transaction)\n  result\nrescue\n  e in FDBC.Error ->\n    :ok = FDBC.Transaction.on_error(tr, e)\n    transact(tr, fun)\nend\n```\n\nThis convenience function allows for transactional blocks to be handled like\nso:\n\n```elixir\ndb = FDBC.Database.create()\nFDBC.transact(db, fn tr ->\n  :ok = Transaction.set(tr, \"foo\", \"bar\")\nend)\n```\n\nIn reality the `FDBC.transact/3` is more versatile than this and can be\nconsulted for further details.","ref":"FDBC.Transaction.html"},{"type":"module","title":"Futures - FDBC.Transaction","doc":"This library unlike [upstream implementations](https://apple.github.io/foundationdb/developer-guide.html#developer-guide-programming-with-futures)\ndoes not support implicit asynchronicity, it must be explicity used.\n\nThere are two ways in which to achieve this, using `Task` or by using the\n`async_*` variants along with `FDBC.Future`.\n\nUsing `Task` the example in the link above can be achieved like so:\n\n```elixir\ntasks = [\n  Task.async(fn -> FDBC.Transaction.get(tr, \"A\") end),\n  Task.async(fn -> FDBC.Transaction.get(tr, \"B\") end),\n]\nresult = Task.await_many(tasks) |> Enum.reduce(0, fn x, acc -> acc <> x end)\nIO.inspect(result)\n```\n\nUsing the `async_*` variants it can be achieved with the following:\n\n```elixir\nfutures = [\n  FDBC.Transaction.async_get(tr, \"A\"),\n  FDBC.Transaction.async_get(tr, \"B\"),\n]\nresult = FDBC.Futures.await_many(futures) |> Enum.reduce(0, fn x, acc -> acc <> x end)\nIO.inspect(result)\n```\n\nThe main difference with the two approaches is that `Task` will spawn each\noperation in a new process while `FDBC.Future` will operate within the\ncalling process.","ref":"FDBC.Transaction.html#module-futures"},{"type":"module","title":"Watches - FDBC.Transaction","doc":"It is possible to [watch keys](https://apple.github.io/foundationdb/developer-guide.html#watches) for\nvalue changes. The most obvious way to handle this is via a `GenServer` like\nimplementation:\n\n```elixir\ndefmodule Watcher do\n  use GenServer\n\n  alias FDBC.{Database, Future, Transaction}\n\n  def start_link(opts) do\n    GenServer.start_link(__MODULE__, :ok, opts)\n  end\n\n  def watch(key) do\n    GenServer.call(__MODULE__, {:watch, key})\n  end\n\n  def init(_opts) do\n    {:ok, %{tasks: %{}}}\n  end\n\n  def handle_call({:watch, key}, _from, state) do\n    {future, task} = start_watch(key)\n    state = put_in(state.tasks[task.ref], key)\n    {:reply, :ok, state}\n  end\n\n  def handle_info({ref, nil}, state) do\n    Process.demonitor(ref, [:flush])\n    {key, state} = pop_in(state.tasks[ref])\n    IO.puts(\"Value for key #{inspect(key)} has changed\")\n    task = start_watch(key)\n    state = put_in(state.tasks[task.ref], key)\n    {:noreply, state}\n  end\n\n  def handle_info({:DOWN, ref, _, _, reason}, state) do\n    {_key, state} = pop_in(state.tasks[ref])\n    # Should check reason to see if watcher was cancelled or errored...\n    {:noreply, state}\n  end\n\n  defp start_watch(key) do\n    tr = Database.create() |> Transaction.create()\n    future = Transaction.watch(tr, key)\n    :ok = Transaction.commit(tr)\n\n    Task.Supervisor.async_nolink(Example.TaskSupervisor, fn ->\n      Future.resolve(future)\n    end)\n  end\nend\n```\n\nThe above would then need to be added to the application's supervision tree:\n\n```elixir\nchildren = [\n  {Task.Supervisor, name: Example.TaskSupervisor},\n  {Watcher, name: Example.Watcher}\n]\n\nSupervisor.start_link(children, strategy: :one_for_one)\n```","ref":"FDBC.Transaction.html#module-watches"},{"type":"function","title":"FDBC.Transaction.add_conflict_key/3","doc":"Adds a conflict key to a transaction without performing the associated read or write.\n\nWorks the same way as `add_conflict_range/4` by creating the range on a single key.","ref":"FDBC.Transaction.html#add_conflict_key/3"},{"type":"function","title":"FDBC.Transaction.add_conflict_range/4","doc":"Adds a [conflict range](https://apple.github.io/foundationdb/developer-guide.html#conflict-ranges)\nto a transaction without performing the associated read or write.\n\nIf `:read` is used, this function adds a range of keys to the transaction’s\nread conflict ranges as if you had read the range. As a result, other\ntransactions that write a key in this range could cause the transaction to\nfail with a conflict.\n\nIf `:write` is used, this function adds a range of keys to the transaction’s\nwrite conflict ranges as if you had cleared the range. As a result, other\ntransactions that concurrently read a key in this range could fail with a\nconflict.","ref":"FDBC.Transaction.html#add_conflict_range/4"},{"type":"function","title":"FDBC.Transaction.async_get/3","doc":"The same as `get/3` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get/3"},{"type":"function","title":"FDBC.Transaction.async_get_addresses_for_key/2","doc":"The same as `get_addresses_for_key/2` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_addresses_for_key/2"},{"type":"function","title":"FDBC.Transaction.async_get_approximate_size/1","doc":"The same as `get_approximate_size/1` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_approximate_size/1"},{"type":"function","title":"FDBC.Transaction.async_get_estimated_range_size/3","doc":"The same as `get_estimated_range_size/3` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_estimated_range_size/3"},{"type":"function","title":"FDBC.Transaction.async_get_key/3","doc":"The same as `get_key/3` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_key/3"},{"type":"function","title":"FDBC.Transaction.async_get_read_version/1","doc":"The same as `get_read_version/1` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_read_version/1"},{"type":"function","title":"FDBC.Transaction.async_get_tag_throttled_duration/1","doc":"The same as `get_tag_throttled_duration/1` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_tag_throttled_duration/1"},{"type":"function","title":"FDBC.Transaction.async_get_total_cost/1","doc":"The same as `get_total_cost/1` except it returns the unresolved future.","ref":"FDBC.Transaction.html#async_get_total_cost/1"},{"type":"function","title":"FDBC.Transaction.async_get_versionstamp/1","doc":"Returns a future that will resolve to the versionstamp used by the transaction.\n\nThe underlying future will be ready only after the successful completion of a\ncall to `commit/1`. Read-only transactions do not modify the database when\ncommitted and will result in the underlying future completing with an error.\nKeep in mind that a transaction which reads keys and then sets them to their\ncurrent values may be optimized to a read-only transaction.\n\n> #### Warning {: .warning}\n> It must be called before `commit/1` but resolved after it.","ref":"FDBC.Transaction.html#async_get_versionstamp/1"},{"type":"function","title":"FDBC.Transaction.atomic_op/4","doc":"Perform a mutation as an atomic operation against the database.\n\nTo be more specific an atomic operation modifies the database snapshot\nrepresented by transaction to perform the operation indicated by `op` with\noperand `param` to the value stored by the given `key`.\n\nAn atomic operation is a single database command that carries out several\nlogical steps: reading the value of a key, performing a transformation on\nthat value, and writing the result. Different atomic operations perform\ndifferent transformations. Like other database operations, an atomic\noperation is used within a transaction; however, its use within a transaction\nwill not cause the transaction to conflict.\n\nAtomic operations do not expose the current value of the key to the client\nbut simply send the database the transformation to apply. In regard to\nconflict checking, an atomic operation is equivalent to a write without a\nread. It can only cause other transactions performing reads of the key to\nconflict.\n\nBy combining these logical steps into a single, read-free operation,\nFoundationDB can guarantee that the transaction will not conflict due to the\noperation. This makes atomic operations ideal for operating on keys that are\nfrequently modified. A common example is the use of a key-value pair as a\ncounter.","ref":"FDBC.Transaction.html#atomic_op/4"},{"type":"function","title":"Mutations - FDBC.Transaction.atomic_op/4","doc":"* `:add` - Performs an addition of little-endian integers. If the existing\n    value in the database is not present or shorter than `param`, it is first\n    extended to the length of `param` with zero bytes. If `param` is shorter\n    than the existing value in the database, the existing value is truncated\n    to match the length of `param`. The integers to be added must be stored\n    in a little-endian representation. They can be signed in two's complement\n    representation or unsigned. You can add to an integer at a known offset\n    in the value by prepending the appropriate number of zero bytes to\n    `param` and padding with zero bytes to match the length of the value.\n    However, this offset technique requires that you know the addition will\n    not cause the integer field within the value to overflow.\n\n  * `:append_if_fits` - Appends `param` to the end of the existing value\n    already in the database at the given key (or creates the key and sets the\n    value to `param` if the key is empty). This will only append the value if\n    the final concatenated value size is less than or equal to the maximum\n    value size. WARNING: No error is surfaced back to the user if the final\n    value is too large because the mutation will not be applied until after\n    the transaction has been committed. Therefore, it is only safe to use\n    this mutation type if one can guarantee that one will keep the total\n    value size under the maximum size.\n\n  * `:bit_and` - Performs a bitwise `and` operation. If the existing value in\n    the database is not present, then `param` is stored in the database. If\n    the existing value in the database is shorter than `param`, it is first\n    extended to the length of `param` with zero bytes. If `param` is shorter\n    than the existing value in the database, the existing value is truncated\n    to match the length of `param`.\n\n  * `:bit_or` - Performs a bitwise `or` operation. If the existing value in\n    the database is not present or shorter than `param`, it is first extended\n    to the length of `param` with zero bytes. If `param` is shorter than the\n    existing value in the database, the existing value is truncated to match\n    the length of `param`.\n\n  * `:bit_xor` - Performs a bitwise `xor` operation. If the existing value in\n    the database is not present or shorter than `param`, it is first extended\n    to the length of `param` with zero bytes. If `param` is shorter than the\n    existing value in the database, the existing value is truncated to match\n    the length of `param`.\n\n  * `:byte_max` - Performs lexicographic comparison of byte strings. If the\n    existing value in the database is not present, then `param` is stored.\n    Otherwise the larger of the two values is then stored in the database.\n\n  * `:byte_min` - Performs lexicographic comparison of byte strings. If the\n    existing value in the database is not present, then `param` is stored.\n    Otherwise the smaller of the two values is then stored in the database.\n\n  * `:compare_and_clear` - Performs an atomic `compare and clear` operation.\n    If the existing value in the database is equal to the given value, then\n    given key is cleared.\n\n  * `:max` - Performs a little-endian comparison of byte strings. If the\n    existing value in the database is not present or shorter than `param`, it\n    is first extended to the length of `param` with zero bytes. If `param` is\n    shorter than the existing value in the database, the existing value is\n    truncated to match the length of `param`. The larger of the two values is\n    then stored in the database.\n\n  * `:min` - Performs a little-endian comparison of byte strings. If the\n    existing value in the database is not present, then `param` is stored in\n    the database. If the existing value in the database is shorter than\n    `param`, it is first extended to the length of `param` with zero bytes.\n    If `param` is shorter than the existing value in the database, the\n    existing value is truncated to match the length of `param`. The smaller\n    of the two values is then stored in the database.\n\n  * `:set_versionstamped_key` - Transforms `key` using a versionstamp for the\n    transaction. Sets the transformed key in the database to `param`. The key\n    is transformed by removing the final four bytes from the key and reading\n    those as a little-Endian 32-bit integer to get a position `pos`. The 10\n    bytes of the key from `pos` to `pos + 10` are replaced with the\n    versionstamp of the transaction used. The first byte of the key is\n    position 0. A versionstamp is a 10 byte, unique, monotonically (but not\n    sequentially) increasing value for each committed transaction. The first\n    8 bytes are the committed version of the database (serialized in\n    big-Endian order). The last 2 bytes are monotonic in the serialization\n    order for transactions.\n\n  * `:set_versionstamped_value` - Transforms `param` using a versionstamp for the\n    transaction. Sets the `key` given to the transformed `param`. The\n    parameter is transformed by removing the final four bytes from `param`\n    and reading those as a little-Endian 32-bit integer to get a position\n    `pos`. The 10 bytes of the parameter from `pos` to `pos + 10` are\n    replaced with the versionstamp of the transaction used. The first byte of\n    the parameter is position 0. A versionstamp is a 10 byte, unique,\n    monotonically (but not sequentially) increasing value for each committed\n    transaction. The first 8 bytes are the committed version of the database\n    (serialized in big-Endian order). The last 2 bytes are monotonic in the\n    serialization order for transactions.","ref":"FDBC.Transaction.html#atomic_op/4-mutations"},{"type":"function","title":"FDBC.Transaction.cancel/1","doc":"Cancels the transaction.\n\nAll pending or future uses of the transaction will raise a transaction\ncancelled exception. The transaction can be used again after it is `reset/1`.","ref":"FDBC.Transaction.html#cancel/1"},{"type":"function","title":"FDBC.Transaction.change/2","doc":"Change the options on the transaction after its initial creation.\n\nSee `create/2` for the list of options.","ref":"FDBC.Transaction.html#change/2"},{"type":"function","title":"FDBC.Transaction.clear/3","doc":"Clear the given key from the database.\n\nModify the database snapshot represented by transaction to remove the given\nkey from the database. If the key was not previously present in the database,\nthere is no effect.","ref":"FDBC.Transaction.html#clear/3"},{"type":"function","title":"Options - FDBC.Transaction.clear/3","doc":"* `:no_write_conflict_range` - (true) The operation will not generate a\n    write conflict range. As a result, other transactions which read the\n    key(s) being modified by the next write will not conflict with this\n    transaction. NOTE: This is equivalent to setting\n    `:next_write_no_write_conflict_range` on the transaction followed by\n    calling this function.","ref":"FDBC.Transaction.html#clear/3-options"},{"type":"function","title":"FDBC.Transaction.clear_range/4","doc":"Clear the given range from the database.\n\nModify the database snapshot represented by transaction to remove all keys\n(if any) which are lexicographically greater than or equal to the given begin\nkey and lexicographically less than the given end_key.\n\nRange clears are efficient with FoundationDB – clearing large amounts of data\nwill be fast. However, this will not immediately free up disk - data for the\ndeleted range is cleaned up in the background. For purposes of computing the\ntransaction size, only the begin and end keys of a clear range are counted.\nThe size of the data stored in the range does not count against the\ntransaction size limit.","ref":"FDBC.Transaction.html#clear_range/4"},{"type":"function","title":"Options - FDBC.Transaction.clear_range/4","doc":"* `:no_write_conflict_range` - (true) The operation will not generate a\n    write conflict range. As a result, other transactions which read the\n    key(s) being modified by the next write will not conflict with this\n    transaction. NOTE: This is equivalent to setting\n    `:next_write_no_write_conflict_range` on the transaction followed by\n    calling this function.","ref":"FDBC.Transaction.html#clear_range/4-options"},{"type":"function","title":"FDBC.Transaction.clear_starts_with/3","doc":"Clear all keys starting with the given prefix from the database.\n\nCalls `clear_range/4` under the hood turning the `prefix` into a start/stop\npair that will clear a range of keys that is equivalent to clearing by\nprefix.","ref":"FDBC.Transaction.html#clear_starts_with/3"},{"type":"function","title":"Options - FDBC.Transaction.clear_starts_with/3","doc":"* `:no_write_conflict_range` - (true) The operation will not generate a\n    write conflict range. As a result, other transactions which read the\n    key(s) being modified by the next write will not conflict with this\n    transaction. NOTE: This is equivalent to setting\n    `:next_write_no_write_conflict_range` on the transaction followed by\n    calling this function.","ref":"FDBC.Transaction.html#clear_starts_with/3-options"},{"type":"function","title":"FDBC.Transaction.commit/1","doc":"Attempts to commit the transaction to the database.\n\nThe commit may or may not succeed - in particular, if a conflicting\ntransaction previously committed, then the commit must fail in order to\npreserve transactional isolation. If the commit does succeed, the transaction\nis durably committed to the database and all subsequently started\ntransactions will observe its effects.","ref":"FDBC.Transaction.html#commit/1"},{"type":"function","title":"FDBC.Transaction.create/2","doc":"Creates a new transaction on the given database or tenant.\n\nIt is valid to pass a transaction through this function in order to modify\nits options.","ref":"FDBC.Transaction.html#create/2"},{"type":"function","title":"Options - FDBC.Transaction.create/2","doc":"* `:access_system_keys` - (true) Allows this transaction to read and modify\n    system keys (those that start with the byte 0xFF). Implies `:raw_access`.\n\n  * `:authorization_token` - (true) Attach given authorization token to the\n    transaction such that subsequent tenant-aware requests are authorized.\n\n  * `:auto_throttle_tag` - (binary) Adds a tag to the transaction that can be\n    used to apply manual or automatic targeted throttling. At most 5 tags can\n    be set on a transaction.\n\n  * `:bypass_storage_quota` - (true) Allows this transaction to bypass\n    storage quota enforcement. Should only be used for transactions that\n    directly or indirectly decrease the size of the tenant group's data.\n\n  * `:bypass_unreadable` - (true) Allows `get` operations to read from\n    sections of keyspace that have become unreadable because of versionstamp\n    operations. These reads will view versionstamp operations as if they were\n    set operations that did not fill in the versionstamp.\n\n  * `:causal_read_disable` - (true) Disable causal reads.\n\n  * `:causal_read_risky` - (true) The read version will be committed, and\n    usually will be the latest committed, but might not be the latest\n    committed in the event of a simultaneous fault and misbehaving clock.\n\n  * `:causal_write_risky` - (true) The transaction, if not self-conflicting,\n    may be committed a second time after commit succeeds, in the event of a\n    fault.\n\n  * `:consistency_check_required_replicas` - (integer) Specifies the number\n    of storage server replica results that the load balancer needs to compare\n    when `:enable_replica_consistency_check` option is set.\n\n  * `:debug_transaction_identifier` - (binary) Sets a client provided\n    identifier for the transaction that will be used in scenarios like\n    tracing or profiling. Client trace logging or transaction profiling must\n    be separately enabled.\n\n  * `:enable_replica_consistency_check` - (true) Enables replica consistency\n    check, which compares the results returned by storage server replicas (as\n    many as specified by co  nsistency_check_required_replicas option) for a\n    given read request, in client-side load balancer.\n\n  * `:expensive_clear_cost_estimation_enable` - (true) Asks storage servers\n    for how many bytes a clear key range contains. Otherwise uses the\n    location cache to roughly estimate this.\n\n  * `:first_in_batch` - (true) No other transactions will be applied before\n    this transaction within the same commit version.\n\n  * `:lock_aware` - (true) The transaction can read and write to locked\n    databases, and is responsible for checking that it took the lock.\n\n  * `:log_transaction` - (true) Enables tracing for this transaction and logs\n    results to the client trace logs. The `:debug_transaction_identifier`\n    option must be set before using this option, and client trace logging\n    must be enabled to get log output.\n\n  * `:max_retry_delay` - (integer) Set the maximum amount of backoff delay\n    incurred in the call to `on_error` if the error is retryable. If the\n    maximum retry delay is less than the current retry delay of the\n    transaction, then the current retry delay will be clamped to the maximum\n    retry delay. Prior to API version 610, like all other transaction\n    options, the maximum retry delay must be reset after a call to\n    `on_error`. If the API version is 610 or greater, the retry limit is not\n    reset after an `on_error` call. Note that at all API versions, it is\n    safe and legal to set the maximum retry delay each time the transaction\n    begins, so most code written assuming the older behavior can be upgraded\n    to the newer behavior without requiring any modification, and the caller\n    is not required to implement special logic in retry loops to only\n    conditionally set this option. Defaults to `1000`.\n\n  * `:next_write_no_write_conflict_range` - (true) The next write performed\n    on this transaction will not generate a write conflict range. As a\n    result, other transactions which read the key(s) being modified by the\n    next write will not conflict with this transaction. Care needs to be\n    taken when using this option on a transaction that is shared between\n    multiple threads. When setting this option, write conflict ranges will be\n    disabled on the next write operation, regardless of what thread it is on.\n\n  * `:priority` - (atom) Set the priority for the transaction. Valid values\n    are `:immediate` and `:batch`. Where `:immediate` specifies that this\n    transaction should be treated as highest priority and that lower priority\n    transactions should block behind this one. Use is discouraged outside of\n    low-level tools. While `:batch` specifies that this transaction should be\n    treated as low priority and that default priority transactions will be\n    processed first. Batch priority transactions will also be throttled at\n    load levels smaller than for other types of transactions and may be fully\n    cut off in the event of machine failures. Useful for doing batch work\n    simultaneously with latency-sensitive work.\n\n  * `:raw_access` - (true) Allows this transaction to access the raw\n    key-space when tenant mode is on.\n\n  * `:read_lock_aware` - (true) The transaction can read from locked\n    databases.\n\n  * `:read_priority` - (atom) Set the priority for subsequent read requests\n    in this transaction. Valid values are `:low`, `:normal`, and `:high`.\n    Defaults to `:normal`.\n\n  * `:read_server_side_cache` - (boolean) Whether the storage server should\n    cache disk blocks needed for subsequent read requests in this\n    transaction. Defaults to `true`.\n\n  * `:read_system_keys` - (true) Allows this transaction to read system keys\n    (those that start with the byte 0xFF). Implies `:raw_access`.\n\n  * `:read_your_writes_disabled` - (true) Reads performed by a transaction\n    will not see any prior mutations that occurred in that transaction,\n    instead seeing the value which was in the database at the transaction's\n    read version. This option may provide a small performance benefit for the\n    client, but also disables a number of client-side optimizations which are\n    beneficial for transactions which tend to read and write the same keys\n    within a single transaction. It is an error to set this option after\n    performing any reads or writes on the transaction.\n\n  * `:report_conflicting_keys` - (true) The transaction can retrieve keys\n    that are conflicting with other transactions.\n\n  * `:retry_limit` - (integer) Set a maximum number of retries after which\n    additional calls to `on_error` will throw the most recently seen error\n    code. If set to `-1`, will disable the retry limit. Prior to API version\n    610, like all other transaction options, the retry limit must be reset\n    after a call to `on_error`. If the API version is 610 or greater, the\n    retry limit is not reset after an `on_error` call. Note that at all API\n    versions, it is safe and legal to set the retry limit each time the trans\n    action begins, so most code written assuming the older behavior can be\n    upgraded to the newer behavior without requiring any modification, and\n    the caller is not required to implement special logic in retry loops to\n    only conditionally set this option.\n\n  * `:server_request_tracing` - (true) Sets an identifier for server tracing\n    of this transaction. When committed, this identifier triggers logging\n    when each part of the transaction authority encounters it, which is\n    helpful in diagnosing slowness in misbehaving clusters. The identifier is\n    randomly generated. When there is also a `:debug_transaction_identifier`,\n    both IDs are logged together.\n\n  * `:size_limit` - (integer) Set the transaction size limit in bytes. The\n    size is calculated by combining the sizes of all keys and values written\n    or mutated, all key ranges cleared, and all read and write conflict\n    ranges. (In other words, it includes the total size of all data included\n    in the request to the cluster to commit the transaction.) Large\n    transactions can cause performance problems on FoundationDB clusters, so\n    setting this limit to a smaller value than the default can help prevent\n    the client from accidentally degrading the cluster's performance. This\n    value must be at least 32 and cannot be set to higher than 10,000,000,\n    the default transaction size limit.\n\n  * `:snapshot_ryw` - (boolean) Allow snapshot read operations will see the\n    results of writes done in the same transaction. Defaults to `true`.\n\n  * `:span_parent` - (binary) Adds a parent to the Span of this transaction.\n    Used for transaction tracing. A span can be identified with a 33 bytes\n    serialized binary format which consists of: 8 bytes protocol version,\n    e.g. ``0x0FDBC00B073000000LL`` in little-endian format, 16 bytes trace id,\n    8 bytes span id, 1 byte set to 1 if sampling is enabled.\n\n  * `:special_key_space_enable_writes` - (true) By default, users are not\n    allowed to write to special keys. Enable this option will implicitly\n    enable all options required to achieve the configuration change.\n\n  * `:special_key_space_relaxed` - (true) By default, the special key space\n    will only allow users to read from exactly one module (a subspace in the\n    special key space). Use this option to allow reading from zero or more\n    modules. Users who set this option should be prepared for new modules,\n    which may have different behaviors than the modules they're currently\n    reading. For example, a new module might block or return an error.\n\n  * `:tag` - (binary) Adds a tag to the transaction that can be used to apply\n    manual targeted throttling. At most 5 tags can be set on a transaction.\n\n  * `:timeout` - (integer) Set a timeout in milliseconds which, when elapsed,\n    will cause the transaction automatically to be cancelled. If set to `0`,\n    will disable all timeouts. All pending and any future uses of the\n    transaction will throw a  n exception. The transaction can be used again\n    after it is reset. Prior to API version 610, like all other transaction\n    options, the timeout must be reset after a call to `on_error`. If the\n    API version is 610 or greater, the timeout is not reset after an\n    `on_error` call. This allows the user to specify a longer timeout on\n    specific transactions than the default timeout specified through the\n    `:transaction_timeout` database option without the shorter database\n    timeout cancelling transactions that encounter a retryable error. Note\n    that at all API versions, it is safe and legal to set the timeout each\n    time the transaction begins, so most code written assum  ing the older\n    behavior can be upgraded to the newer behavior without requiring any\n    modification, and the caller is not required to i  mplement special logic\n    in retry loops to only conditionally set this option.\n\n  * `:transaction_logging_max_field_length` - (integer) Sets the maximum\n    escaped length of key and value fields to be logged to the trace file via the\n    `:log_transaction` option, after which the field will be truncated. A\n    negative value disables truncation.\n\n  * `:use_grv_cache` - (boolean) Allows this transaction to use cached GRV\n    from the database context. Upon first usage, starts a background updater\n    to periodically update the cache to avoid stale read versions. The\n    disable_client_bypass option must also be set. Defaults to `false`.\n\n  * `:used_during_commit_protection_disable` - (true) By default, operations\n    that are performed on a transaction while it is being committed will not\n    only fail themselves, but the  y will attempt to fail other in-flight\n    operations (such as the commit) as well. This behavior is intended to\n    help developers discove  r situations where operations could be\n    unintentionally executed after the transaction has been reset. Setting\n    this option removes th  at protection, causing only the offending\n    operation to fail.","ref":"FDBC.Transaction.html#create/2-options"},{"type":"function","title":"FDBC.Transaction.get/3","doc":"Get a value from the database.\n\nReads a value from the database snapshot represented by the transaction.","ref":"FDBC.Transaction.html#get/3"},{"type":"function","title":"Options - FDBC.Transaction.get/3","doc":"* `:snapshot` - (boolean) Perform the get as a [snapshot read](https://apple.github.io/foundationdb/api-c.html#snapshots).","ref":"FDBC.Transaction.html#get/3-options"},{"type":"function","title":"FDBC.Transaction.get_addresses_for_key/2","doc":"Returns the storage server adressess storing the given key.\n\nReturns a list of public network addresses as strings, one for each of the\nstorage servers responsible for storing key_name and its associated value.","ref":"FDBC.Transaction.html#get_addresses_for_key/2"},{"type":"function","title":"FDBC.Transaction.get_approximate_size/1","doc":"Returns the approximate transaction size so far.\n\nThis is a summation of the estimated size of mutations, read conflict ranges,\nand write conflict ranges.","ref":"FDBC.Transaction.html#get_approximate_size/1"},{"type":"function","title":"FDBC.Transaction.get_committed_version/1","doc":"Returns the database version number for the commited transaction.\n\nRetrieves the database version number at which a given transaction was\ncommitted. `commit/1` must have been called on transaction. Read-only\ntransactions do not modify the database when committed and will have a\ncommitted version of -1. Keep in mind that a transaction which reads keys and\nthen sets them to their current values may be optimized to a read-only\ntransaction.\n\nNote that database versions are not necessarily unique to a given transaction\nand so cannot be used to determine in what order two transactions completed.\nThe only use for this function is to manually enforce causal consistency when\ncalling `set_read_version/2` on another subsequent transaction.","ref":"FDBC.Transaction.html#get_committed_version/1"},{"type":"function","title":"FDBC.Transaction.get_estimated_range_size/3","doc":"Returns an estimated byte size of the key range.\n\n> #### Note {: .info}\n> The estimated size is calculated based on the sampling done by FDB server.\n  The sampling algorithm works roughly in this way: the larger the key-value\n  pair is, the more likely it would be sampled and the more accurate its\n  sampled size would be. And due to that reason it is recommended to use this\n  API to query against large ranges for accuracy considerations. For a rough\n  reference, if the returned size is larger than 3MB, one can consider the\n  size to be accurate.","ref":"FDBC.Transaction.html#get_estimated_range_size/3"},{"type":"function","title":"FDBC.Transaction.get_key/3","doc":"Returns the first key in the database that matches the given key selector.","ref":"FDBC.Transaction.html#get_key/3"},{"type":"function","title":"Options - FDBC.Transaction.get_key/3","doc":"* `:snapshot` - (boolean) Perform the get as a [snapshot read](https://apple.github.io/foundationdb/api-c.html#snapshots).","ref":"FDBC.Transaction.html#get_key/3-options"},{"type":"function","title":"FDBC.Transaction.get_metadata_version/1","doc":"Returns the metadata version.\n\nThe metadata version key `\\xff/metadataVersion` is a key intended to help\nlayers deal with hot keys. The value of this key is sent to clients along\nwith the read version from the proxy, so a client can read its value without\ncommunicating with a storage server.\n\nIt is stored as a versionstamp, and can be `nil` if its yet to be utilised.","ref":"FDBC.Transaction.html#get_metadata_version/1"},{"type":"function","title":"FDBC.Transaction.get_range/4","doc":"Returns all the key-value pairs for the given range.","ref":"FDBC.Transaction.html#get_range/4"},{"type":"function","title":"Options - FDBC.Transaction.get_range/4","doc":"* `:limit` - (integer) Indicates the maximum number of key-value pairs to\n    return.\n\n  * `:mode` - (atom) The mode in which to return the data to the caller.\n    Defaults to `:iterator`.\n\n    * `:exact` - The client has passed a specific row limit and wants that\n      many rows delivered in a single batch. Because of iterator operation in\n      client drivers make request batches transparent to the user, consider\n      `:want_all` instead. A row `:limit` must be specified if this mode is\n      used.\n\n    * `:iterator` - The client doesn't know how much of the range it is\n      likely to used and wants different performance concerns to be balanced.\n      Only a small portion of data is transferred to the client initially (in\n      order to minimize costs if the client doesn't read the entire range),\n      and as the caller iterates over more items in the range larger batches\n      will be transferred in order to minimize latency. After enough\n      iterations, the iterator mode will eventually reach the same byte limit\n      as `:want_all`.\n\n    * `:large` - Transfer data in batches large enough to be, in a\n      high-concurrency environment, nearly as efficient as possible. If the\n      client stops iteration early, some disk and network bandwidth may be\n      wasted. The batch size may still be too small to allow a single client\n      to get high throughput from the database, so if that is what you need\n      consider the `:serial` instead.\n\n    * `:medium` - Transfer data in batches sized in between small and large.\n\n    * `:serial` - Transfer data in batches large enough that an individual\n      client can get reasonable read bandwidth from the database. If the\n      client stops iteration early, considerable disk and network bandwidth\n      may be wasted.\n\n    * `:small` - Transfer data in batches small enough to not be much more\n      expensive than reading individual rows, to minimize cost if iteration\n      stops early.\n\n    * `:want_all` - Client intends to consume the entire range and would like\n      it all transferred as early as possible.\n\n  * `:reverse` - (boolean) The key-value pairs will be returned in reverse\n    lexicographical order beginning at the end of the range. Reading ranges\n    in reverse is supported natively by the database and should have minimal\n    extra cost.\n\n  * `:snapshot` - (boolean) Perform the get as a [snapshot read](https://apple.github.io/foundationdb/api-c.html#snapshots).","ref":"FDBC.Transaction.html#get_range/4-options"},{"type":"function","title":"FDBC.Transaction.get_range_split_points/4","doc":"Returns a list of keys that can split the given range into roughly equally sized chunks based on chunk `size`.\n\nThe returned split points contain the start key and end key of the given range.","ref":"FDBC.Transaction.html#get_range_split_points/4"},{"type":"function","title":"FDBC.Transaction.get_read_version/1","doc":"Returns the transaction snapshot read version.\n\nThe transaction obtains a snapshot read version automatically at the time of\nthe first call to `get_*()` (including this one) and (unless causal\nconsistency has been deliberately compromised by transaction options) is\nguaranteed to represent all transactions which were reported committed before\nthat call.","ref":"FDBC.Transaction.html#get_read_version/1"},{"type":"function","title":"FDBC.Transaction.get_starts_with/3","doc":"Returns all the key-value pairs the start with the given prefix.\n\nThis function calls `get_range/4` and therefore supports the same options as\nit.","ref":"FDBC.Transaction.html#get_starts_with/3"},{"type":"function","title":"FDBC.Transaction.get_tag_throttled_duration/1","doc":"Returns the time in seconds that the transaction was throttled by the tag\nthrottler.","ref":"FDBC.Transaction.html#get_tag_throttled_duration/1"},{"type":"function","title":"FDBC.Transaction.get_total_cost/1","doc":"Returns the cost of the transaction so far in bytes.\n\nThe cost is computed by the tag throttler, and used for tag throttling if\nthroughput quotas are specified.","ref":"FDBC.Transaction.html#get_total_cost/1"},{"type":"type","title":"FDBC.Transaction.mutation/0","doc":"","ref":"FDBC.Transaction.html#t:mutation/0"},{"type":"function","title":"FDBC.Transaction.on_error/2","doc":"Implements the recommended retry and backoff behavior for a transaction.\n\nThis function knows which of the error codes generated by other\n`FDBC.Transaction` functions represent temporary error conditions and which\nrepresent application errors that should be handled by the application. It\nalso implements an exponential backoff strategy to avoid swamping the\ndatabase cluster with excessive retries when there is a high level of\nconflict between transactions.","ref":"FDBC.Transaction.html#on_error/2"},{"type":"function","title":"FDBC.Transaction.reset/1","doc":"Reset the transaction to its initial state.\n\nIt is not necessary to call `reset/1` when handling an error with\n`on_error/2` since the transaction has already been reset.","ref":"FDBC.Transaction.html#reset/1"},{"type":"function","title":"FDBC.Transaction.set/4","doc":"Set the value for a given key.","ref":"FDBC.Transaction.html#set/4"},{"type":"function","title":"Options - FDBC.Transaction.set/4","doc":"* `:no_write_conflict_range` - (true) The operation will not generate a\n    write conflict range. As a result, other transactions which read the\n    key(s) being modified by the next write will not conflict with this\n    transaction. NOTE: This is equivalent to setting\n    `:next_write_no_write_conflict_range` on the transaction followed by\n    calling this function.","ref":"FDBC.Transaction.html#set/4-options"},{"type":"function","title":"FDBC.Transaction.set_metadata_version/1","doc":"Sets the metadata version.\n\nIt takes no value as the database is responsible for setting it.","ref":"FDBC.Transaction.html#set_metadata_version/1"},{"type":"function","title":"FDBC.Transaction.set_read_version/2","doc":"Sets the snapshot read version.\n\nThis is not needed in simple cases. If the given version is too old,\nsubsequent reads will fail with error code 'transaction_too_old'; if it is too\nnew, subsequent reads may be delayed indefinitely and/or fail with\nerror code 'future_version'. If any of get_*() have been called on this\ntransaction already, the result is undefined.","ref":"FDBC.Transaction.html#set_read_version/2"},{"type":"function","title":"FDBC.Transaction.stream/4","doc":"Streams all the key-value pairs for the given range.\n\nThis function supports the same options as `get_range/4`.","ref":"FDBC.Transaction.html#stream/4"},{"type":"function","title":"FDBC.Transaction.stream_starts_with/3","doc":"Stream all the key-value pairs the start with the given prefix.\n\nThis function calls `get_range/4` and therefore supports the same options as\nit.","ref":"FDBC.Transaction.html#stream_starts_with/3"},{"type":"type","title":"FDBC.Transaction.t/0","doc":"","ref":"FDBC.Transaction.html#t:t/0"},{"type":"function","title":"FDBC.Transaction.watch/2","doc":"Watch for a change on the given key's value.\n\nA watch’s behavior is relative to the transaction that created it. A watch\nwill report a change in relation to the key’s value as readable by that\ntransaction. The initial value used for comparison is either that of the\ntransaction’s read version or the value as modified by the transaction itself\nprior to the creation of the watch. If the value changes and then changes\nback to its initial value, the watch might not report the change.\n\nUntil the transaction that created it has been committed, a watch will not\nreport changes made by other transactions. In contrast, a watch will\nimmediately report changes made by the transaction itself. Watches cannot be\ncreated if the transaction has set the `:read_your_writes_disabled`\ntransaction option, and an attempt to do so will return an watches_disabled\nerror.\n\nIf the transaction used to create a watch encounters an error during commit,\nthen the watch will be set with that error. A transaction whose commit result\nis unknown will set all of its watches with the commit_unknown_result error.\nIf an uncommitted transaction is reset or destroyed, then any watches it\ncreated will be set with the transaction_cancelled error.\n\nReturns an `FDBC.Future` representing an empty value that will be set once\nthe watch has detected a change to the value at the specified key.\n\nBy default, each database connection can have no more than 10,000 watches\nthat have not yet reported a change. When this number is exceeded, an attempt\nto create a watch will return a too_many_watches error. This limit can be\nchanged using the `:max_watches` database option. Because a watch outlives the\ntransaction that creates it, any watch that is no longer needed should be\ncancelled by calling `cancel/1` on its returned future.","ref":"FDBC.Transaction.html#watch/2"},{"type":"module","title":"FDBC.Tuple","doc":"An implementation of upstreams builtin tuple layer. This layer is used for\nencoding keys that are useable by FoundationDB. The encoded key maintains the\nsame sort order as the original tuple: sorted first by the first element,\nthen by the second element, etc. This makes the tuple layer ideal for\nbuilding a variety of higher-level data models.\n\nBelow is an example showing the usage of the tuple layer for handling\npopulation figures for the United States are stored using keys formed from\nthe tuple of state and county:\n\n```elixir\ndefmodule Example do\n  def get_county_populations_in_state(db, state) do\n    FDBC.transact(db, fn tr ->\n      {start, stop} = FDBC.Tuple.range([state])\n      Transaction.get_range(tr, start, stop) |> Enum.map(fn {_, v} -> String.to_integer(v) end)\n    end)\n  end\n\n  def set_county_population(db, state, county, population) do\n    FDBC.transact(db, fn tr ->\n      key = FDBC.Tuple.pack([state, county])\n      Transaction.set(tr, key, Integer.to_string(population))\n    end)\n  end\nend\n```","ref":"FDBC.Tuple.html"},{"type":"module","title":"Upstream Compatibility - FDBC.Tuple","doc":"At the time of writing all [type codes](https://github.com/apple/foundationdb/blob/main/design/tuple.md) are\nimplemented.\n\nAlthough fully complaint, the tuple layer implementation is lossy by default\ndue to Elixir's type system. While not an issue when the key space is\nexclusively managed by this library, this is not the case if the key space is\nshared. This is because types like `t:String.t/0` map directly down to\n`t:binary/0` which is a lossy process at the time of packing.\n\n```elixir\n# A string will encode to the type code of a binary and not of a unicode string.\nassert \"\\x01foobar\\x00\" == Tuple.pack([\"foobar\"])\n\n# This can be countered via a keyword list.\nassert \"\\x02foobar\\x00\" == Tuple.pack([{string: \"foobar\"}])\n```\n\nTo allow interoperability on shared key spaces the `FDBC.Tuple` module\nsupports the `:keyed` option. When used this option changes the tuple from a\nlist into a keyword list, where the keys are used to type the tuple.\n\n```elixir\n# In this case `keyed` will ensure the tuple is a keyword list and raise an exception if not.\nassert \"\\x02foobar\\x00\" == Tuple.pack([{string: \"foobar\"}], keyed: true)\n\n# In this case `keyed` will ensure the tuple is returned as a keyword list.\nassert [{string: \"foobar\"}] == Tuple.unpack(\"\\x02foobar\\x00\", keyed: true)\n```\n\nThe following atoms are used for keying tuple type codes:\n\n* `:null` - `0x00`\n* `:binary` - `0x01`\n* `:string` - `0x02`\n* `:integer` - `0x0B..0x1D`\n* `:float` - `0x20`\n* `:double` - `0x21`\n* `:boolean` - `0x26..0x27`\n* `:uuid` - `0x30`\n* `:versionstamp` - `0x33`","ref":"FDBC.Tuple.html#module-upstream-compatibility"},{"type":"module","title":"User Type Codes - FDBC.Tuple","doc":"It is possible to extend the tuple layer with user type codes which can be\nachieved by implementing the `FDBC.Tuple.Extension` behaviour and supplying\nit to the tuple functions via the `:extension` option. For more information\nand example refer to the documentation in `FDBC.Tuple.Extension`.","ref":"FDBC.Tuple.html#module-user-type-codes"},{"type":"function","title":"FDBC.Tuple.pack/2","doc":"Returns a key encoding the specified tuple.\n\nRaises an `ArgumentError` if more than one incomplete\n`FDBC.Tuple.Versionstamp` is provided.","ref":"FDBC.Tuple.html#pack/2"},{"type":"function","title":"Options - FDBC.Tuple.pack/2","doc":"* `:extension` - a module that implements the behaviour `FDBC.Tuple.Extension`\n    allowing the Tuple layer to be extended with user type codes.\n\n  * `:keyed` - if present, then the tuple is returned as a keyword list, such\n    as `[{string: \"flip\"}, {string: \"flop\"}]`.","ref":"FDBC.Tuple.html#pack/2-options"},{"type":"function","title":"FDBC.Tuple.range/2","doc":"Returns the start and stop keys for a given tuple.\n\nThe start and stop keys allow the retrieval of all tuples of greater length\nthan the provided `prefix`.","ref":"FDBC.Tuple.html#range/2"},{"type":"function","title":"Options - FDBC.Tuple.range/2","doc":"* `:extension` - a module that implements the behaviour `FDBC.Tuple.Extension`\n    allowing the Tuple layer to be extended with user type codes.\n\n  * `:keyed` - if present, then the tuple must be keyedly typed as a\n    keyword list, such as `[{string: \"flip\"}, {string: \"flop\"}]`.","ref":"FDBC.Tuple.html#range/2-options"},{"type":"function","title":"Examples - FDBC.Tuple.range/2","doc":"# Returns start & stop keys which when used in a range query would return\n  # tuples like [\"A\", 2, x], [\"A\", 2, x, y], etc.\n  FDBC.Tuple.range([\"A\", 2])","ref":"FDBC.Tuple.html#range/2-examples"},{"type":"function","title":"FDBC.Tuple.unpack/2","doc":"Returns the tuple that encoded into the given key.","ref":"FDBC.Tuple.html#unpack/2"},{"type":"function","title":"Options - FDBC.Tuple.unpack/2","doc":"* `:extension` - a module that implements the behaviour `FDBC.Tuple.Extension`\n    allowing the Tuple layer to be extended with user type codes.\n\n  * `:keyed` - if present, then the tuple is returned as a keyword list, such\n    as `[{string: \"flip\"}, {string: \"flop\"}]`.","ref":"FDBC.Tuple.html#unpack/2-options"},{"type":"behaviour","title":"FDBC.Tuple.Extension","doc":"A behaviour module for implementing user type codes.\n\nUsing this module makes it possible to extend the `FDBC.Tuple` layer with\n[user type codes](https://github.com/apple/foundationdb/blob/main/design/tuple.md#user-type-codes).\nOnly the following type codes are allowed within the implementation\n`0x40..0x4F` as defined in the specification.\n\nBelow is an example showing how to create an extension that would enable the\ntuple layer to handle `DateTime`:\n\n```elixir\ndefmodule TupleExtensions.DateTime do\n  use FDBC.Tuple.Extension\n\n  @datetime_typecode 0x40\n\n  def decode(<<@datetime_typecode, length::integer-size(8)>> <> encoded) do\n    {value, encoded} =\n      case encoded do\n        < > -> {value, encoded}\n        < > -> {value, <<>>}\n      end\n\n    [timestamp] = Tuple.unpack(value)\n    {:ok, datetime} = DateTime.from_unix(timestamp)\n    {{:datetime, datetime}, encoded}\n  end\n\n  def encode({:datetime, data}) do\n    timestamp = DateTime.to_unix(data)\n    packed = Tuple.pack([timestamp])\n    <<@datetime_typecode>> <> < > <> packed\n  end\n\n  def identify(%DateTime{}) do\n    :datetime\n  end\nend\n```\n\nWhen using extensions the official documentation recommends that they are\nused within labelled directories to prevent type code clashes. With that in\nmind an extended tuple could be utilised like so making use of\n`FDBC.Directory`:\n\n```elixir\nFDBC.transact(db, fn tr ->\n  dir = FDBC.Directory.new() |> FDBC.Directory.open!(tr, [\"example\"], label: \"datetime\")\n  key = FDBC.Tuple.pack([DateTime.utc_now()], extension: TupleExtensions.DateTime)\n  :ok = FDBC.Transaction.set(tr, key, <<>>)\nend)\n```","ref":"FDBC.Tuple.Extension.html"},{"type":"callback","title":"FDBC.Tuple.Extension.decode/1","doc":"Takes the encoded key returning the decoded tuple pair and the tail of the\nkey to still be decoded.","ref":"FDBC.Tuple.Extension.html#c:decode/1"},{"type":"callback","title":"FDBC.Tuple.Extension.encode/1","doc":"Takes the keyed type to encode in the form of a tuple pair, returning its\nencoded representation.","ref":"FDBC.Tuple.Extension.html#c:encode/1"},{"type":"callback","title":"FDBC.Tuple.Extension.identify/1","doc":"Takes the custom type and returns its key.","ref":"FDBC.Tuple.Extension.html#c:identify/1"},{"type":"type","title":"FDBC.Tuple.Extension.keyed/0","doc":"","ref":"FDBC.Tuple.Extension.html#t:keyed/0"},{"type":"module","title":"FDBC.Tuple.Versionstamp","doc":"Used to represent values written by versionstamp operations within the tuple\nlayer. This wraps a binary of length 12 that can be used to represent some\nglobal order of items within the database.\n\nThese versions are composed of two separate components:\n  1. the 10-byte `tr` version.\n  2. the two-byte `user` version.\n\nThe `tr` version is set by the database, and it is used to impose an order\nbetween different transactions. This order is guaranteed to be monotonically\nincreasing over time for a given database. In particular, it imposes an order\nthat is consistent with a serialization order of the database’s transactions.\nIf the client elects to leave the `tr` version as its default value of `nil`,\nthen the Versionstamp is considered 'incomplete'. This will cause the first\n10 bytes of the serialized `FDBC.Tuple.Versionstamp` to be filled in with\ndummy bytes when serialized. When used with operations\n`:set_versionstamped_key` or `:set_versionstamped_value` in\n`FDBC.Transaction.atomic_op/4`, an incomplete version can be used to ensure\nthat a key gets written with the current transaction’s version which can be\nuseful for maintaining append-only data structures within the database. If\nthe `tr` version is set to something that is not `nil`, it should be set to a\nbinary of length 10. In this case, the Versionstamp is considered 'complete'.\nThis is the usual case when one reads a serialized `FDBC.Tuple.Versionstamp`\nfrom the database.\n\nThe `user` version should be specified as an integer, but it must fit within\na two-byte unsigned integer. It is set by the client, and it is used to\nimpose an order between items serialized within a single transaction. If left\nunset, then final two bytes of the serialized `FDBC.Tuple.Versionstamp` are\nfilled in with a default constant value.\n\nThe example belows shows how to serialize an incomplete\n`FDBC.Tuple.Versionstamp`, write it to the database and then read it back:\n\n```elixir\ndefmodule Example do\n  def read_versionstamp(db, subspace) do\n    FDBC.transact(db, fn tr ->\n      {start, stop} = FDBC.Subspace.range(subspace)\n      case FDBC.Transaction.get_range(tr, start, stop, limit: 1) do\n        [] -> nil\n        [{k, _}] -> FDBC.Subspace.unpack(k)\n      end\n    end)\n  end\n\n  def write_versionstamp(db, subspace) do\n    FDBC.transact(db, fn tr ->\n      key = FDBC.Subspace.pack(subspace, [Versionstamp.new()])\n      :ok = FDBC.Transaction.atomic_op(tr, :set_versionstamped_key, key, <<>>)\n      FDBC.Transaction.async_get_versionstamp(tr)\n    end) |> FDBC.Future.resolve()\n  end\nend\n\ndb = FDBC.Database.create()\nsubspace = FDBC.Subspace.new(['prefix'])\n{start, stop} = FDBC.Subspace.range(subspace)\n:ok = FDBC.Transaction.clear_range(db, start, stop)\nversion = Example.write_versionstamp(db, subspace)\nv = Example.read_versionstamp(db, subspace)\nassert v == FDBC.Tuple.Versionstamp(tr=version)\n```","ref":"FDBC.Tuple.Versionstamp.html"},{"type":"function","title":"FDBC.Tuple.Versionstamp.dump/1","doc":"Converts a versiontamp into its binary representation.","ref":"FDBC.Tuple.Versionstamp.html#dump/1"},{"type":"function","title":"FDBC.Tuple.Versionstamp.incomplete?/1","doc":"Checks if a versionstamp is incomplete.","ref":"FDBC.Tuple.Versionstamp.html#incomplete?/1"},{"type":"function","title":"FDBC.Tuple.Versionstamp.load/1","doc":"Loads a versionstamp from its binary representation.","ref":"FDBC.Tuple.Versionstamp.html#load/1"},{"type":"function","title":"FDBC.Tuple.Versionstamp.new/2","doc":"Create a versionstamp.\n\nWhere `tr` is the transaction version that is set by the database and `user`\nis the client specified version used to impose an order between items\nserialized within a single transaction.\n\nIf `tr` is `nil` then an incomplete versionstamp is created.","ref":"FDBC.Tuple.Versionstamp.html#new/2"},{"type":"type","title":"FDBC.Tuple.Versionstamp.t/0","doc":"","ref":"FDBC.Tuple.Versionstamp.html#t:t/0"},{"type":"exception","title":"FDBC.Error","doc":"An exception raised when an error is returned from a call to the FDBC NIF.\n\n`FDBC.Error` exceptions have two fields, `:message` (a `t:String.t/0`) and\n`:code` (a `t:integer`) which are public and can be accessed freely when\nreading `FDBC.Error` exceptions. These values are representative of the\nFoundationDB [error codes](https://apple.github.io/foundationdb/api-error-codes.html).","ref":"FDBC.Error.html"},{"type":"type","title":"FDBC.Error.t/0","doc":"","ref":"FDBC.Error.html#t:t/0"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.36.1"}}